import pytest
from elevator_traffic_analysis.building import Building
from elevator_traffic_analysis.floor import Floor
from elevator_traffic_analysis.elevator import Elevator
from elevator_traffic_analysis.passenger import Passenger


class TestBuildingCreation:
    """Tests for Building creation and initialization"""

    def test_building_creation_basic(self):
        """
        Given: Building parameters are provided
        When: A Building object is created
        Then: The building has correct number of floors and elevators with specified capacity
        """
        building = Building(n_floors=5, n_elevators=2, elevator_capacity=10)
        assert len(building.floors) == 5
        assert len(building.elevators) == 2
        assert building.elevator_capacity == 10

    def test_building_creation_single_floor(self):
        """
        Given: Building parameters with single floor
        When: A Building object is created
        Then: The building has exactly one floor and one elevator
        """
        building = Building(n_floors=1, n_elevators=1, elevator_capacity=5)
        assert len(building.floors) == 1
        assert len(building.elevators) == 1

    def test_building_creation_many_floors(self):
        """
        Given: Building parameters with many floors
        When: A Building object is created
        Then: The building has correct number of floors and elevators
        """
        building = Building(n_floors=50, n_elevators=5, elevator_capacity=15)
        assert len(building.floors) == 50
        assert len(building.elevators) == 5
        assert building.elevator_capacity == 15

    def test_building_creation_default_capacity(self):
        """
        Given: Building parameters without explicit capacity
        When: A Building object is created
        Then: The building uses default elevator capacity
        """
        building = Building(n_floors=10, n_elevators=3)
        assert building.elevator_capacity == 10

    def test_building_floors_have_correct_ids(self):
        """
        Given: A building with multiple floors
        When: The building is created
        Then: Floors have sequential IDs starting from 0
        """
        building = Building(n_floors=5, n_elevators=1)
        for i, floor in enumerate(building.floors):
            assert floor.floor_id == f"F__{i}"

    def test_building_elevators_have_correct_ids(self):
        """
        Given: A building with multiple elevators
        When: The building is created
        Then: Elevators have sequential IDs starting from 0
        """
        building = Building(n_floors=5, n_elevators=3)
        for i, elevator in enumerate(building.elevators):
            assert elevator.elevator_id == f"E__{i}"

    def test_building_elevators_reference_building(self):
        """
        Given: A building with elevators
        When: The building is created
        Then: Each elevator has a reference to the building
        """
        building = Building(n_floors=5, n_elevators=2)
        for elevator in building.elevators:
            assert elevator.building is building

    def test_building_elevators_start_at_ground_floor(self):
        """
        Given: A building with multiple elevators
        When: The building is created
        Then: All elevators start at ground floor (floor 0)
        """
        building = Building(n_floors=10, n_elevators=3)
        for elevator in building.elevators:
            assert elevator.current_floor == building.floors[0]


class TestBuildingRepresentation:
    """Tests for Building string representation"""

    def test_repr(self):
        """
        Given: A building with specific parameters
        When: The building's repr is called
        Then: The repr contains building details
        """
        building = Building(n_floors=8, n_elevators=2, elevator_capacity=12)
        repr_str = repr(building)
        assert "Building" in repr_str
        assert "8 floors" in repr_str
        assert "2 elevators" in repr_str
        assert "elevator_capacity=12" in repr_str


class TestBuildingToDictSerialization:
    """Tests for Building to_dict method"""

    def test_to_dict_structure(self):
        """
        Given: A building object
        When: to_dict method is called
        Then: The dictionary has correct structure with type, capacity, floors, and elevators
        """
        building = Building(n_floors=3, n_elevators=1, elevator_capacity=5)
        building_dict = building.to_dict()

        assert building_dict['type'] == 'Building'
        assert building_dict['elevator_capacity'] == 5
        assert 'floors' in building_dict
        assert 'elevators' in building_dict

    def test_to_dict_floors(self):
        """
        Given: A building with multiple floors
        When: to_dict method is called
        Then: All floors are included in the dictionary
        """
        building = Building(n_floors=4, n_elevators=1)
        building_dict = building.to_dict()

        assert len(building_dict['floors']) == 4
        for i in range(4):
            assert f"F__{i}" in building_dict['floors']

    def test_to_dict_elevators(self):
        """
        Given: A building with multiple elevators
        When: to_dict method is called
        Then: All elevators are included in the dictionary
        """
        building = Building(n_floors=5, n_elevators=3)
        building_dict = building.to_dict()

        assert len(building_dict['elevators']) == 3
        for i in range(3):
            assert f"E__{i}" in building_dict['elevators']


class TestBuildingPassengerCreation:
    """Tests for creating and adding passengers"""

    def test_create_and_add_passenger_success(self):
        """
        Given: Valid passenger parameters
        When: create_and_add_passenger is called
        Then: The passenger is created and added to the origin floor
        """
        building = Building(n_floors=10, n_elevators=2)
        success = building.create_and_add_passenger(
            passenger_id="P__1",
            origin_floor_id="F__0",
            destination_floor_id="F__5",
            tick_appeared=0
        )

        assert success is True
        assert len(building.floors[0].waiting_passengers) == 1
        passenger = building.floors[0].waiting_passengers[0]
        assert passenger.passenger_id == "P__1"
        assert passenger.destination_floor == building.floors[5]

    def test_create_and_add_passenger_invalid_origin(self):
        """
        Given: Passenger parameters with invalid origin floor
        When: create_and_add_passenger is called
        Then: The operation returns False
        """
        building = Building(n_floors=5, n_elevators=1)
        success = building.create_and_add_passenger(
            passenger_id="P__1",
            origin_floor_id="F__10",
            destination_floor_id="F__2",
            tick_appeared=0
        )

        assert success is False

    def test_create_and_add_passenger_invalid_destination(self):
        """
        Given: Passenger parameters with invalid destination floor
        When: create_and_add_passenger is called
        Then: The operation returns False
        """
        building = Building(n_floors=5, n_elevators=1)
        success = building.create_and_add_passenger(
            passenger_id="P__1",
            origin_floor_id="F__0",
            destination_floor_id="F__20",
            tick_appeared=0
        )

        assert success is False

    def test_create_and_add_passenger_negative_floor(self):
        """
        Given: Passenger parameters with negative floor number
        When: create_and_add_passenger is called
        Then: The operation returns False
        """
        building = Building(n_floors=10, n_elevators=1)
        success = building.create_and_add_passenger(
            passenger_id="P__1",
            origin_floor_id="F__-1",
            destination_floor_id="F__5",
            tick_appeared=0
        )

        assert success is False

    def test_create_and_add_passenger_malformed_id(self):
        """
        Given: Passenger parameters with malformed floor ID
        When: create_and_add_passenger is called
        Then: The operation returns False
        """
        building = Building(n_floors=10, n_elevators=1)
        success = building.create_and_add_passenger(
            passenger_id="P__1",
            origin_floor_id="invalid",
            destination_floor_id="F__5",
            tick_appeared=0
        )

        assert success is False

    def test_create_multiple_passengers(self):
        """
        Given: Multiple passenger creation requests
        When: create_and_add_passenger is called for each
        Then: All passengers are successfully created and added
        """
        building = Building(n_floors=10, n_elevators=1)

        for i in range(5):
            success = building.create_and_add_passenger(
                passenger_id=f"P__{i}",
                origin_floor_id="F__0",
                destination_floor_id=f"F__{i+1}",
                tick_appeared=i
            )
            assert success is True

        assert len(building.floors[0].waiting_passengers) == 5


class TestBuildingAddWaitingPassenger:
    """Tests for adding waiting passengers"""

    def test_add_waiting_passenger(self):
        """
        Given: A passenger object and a building
        When: add_waiting_passenger is called
        Then: The passenger is added to their origin floor's waiting list
        """
        building = Building(n_floors=5, n_elevators=1)
        origin = building.floors[0]
        destination = building.floors[3]
        passenger = Passenger("P__1", origin, destination, 0)

        building.add_waiting_passenger(passenger)

        assert len(origin.waiting_passengers) == 1
        assert passenger in origin.waiting_passengers


class TestBuildingTickMovement:
    """Tests for elevator movement coordination"""

    def test_tick_movement_idle_elevators(self):
        """
        Given: A building with idle elevators
        When: tick_movement is called
        Then: All elevators remain idle at ground floor
        """
        building = Building(n_floors=5, n_elevators=2)

        # Initially all elevators should be idle
        building.tick_movement()

        for elevator in building.elevators:
            assert elevator.state == "IDLE"
            assert elevator.current_floor == building.floors[0]

    def test_tick_movement_with_target(self):
        """
        Given: An elevator with a target floor set
        When: tick_movement is called multiple times
        Then: The elevator moves towards and reaches the target
        """
        building = Building(n_floors=10, n_elevators=1)
        elevator = building.elevators[0]
        elevator.target_floor = building.floors[5]

        # First tick should start moving (state changes but doesn't move yet)
        building.tick_movement()
        assert elevator.state == "MOVING_UP"
        assert elevator.current_floor == building.floors[0]

        # Second tick actually moves to floor 1
        building.tick_movement()
        assert elevator.current_floor == building.floors[1]

        # Continue ticking until reached target
        for _ in range(4):
            building.tick_movement()

        assert elevator.current_floor == building.floors[5]
        assert elevator.state == "IDLE"


class TestBuildingPassengerHandling:
    """Tests for passenger embark/disembark coordination"""

    def test_embark_waiting_passengers(self):
        """
        Given: Waiting passengers at a floor with an elevator
        When: embark_waiting_passengers is called
        Then: Passengers board the elevator and are removed from waiting list
        """
        building = Building(n_floors=5, n_elevators=1, elevator_capacity=10)
        elevator = building.elevators[0]

        # Add waiting passengers
        building.create_and_add_passenger("P__1", "F__0", "F__3", 0)
        building.create_and_add_passenger("P__2", "F__0", "F__4", 0)

        assert len(building.floors[0].waiting_passengers) == 2

        # Embark passengers
        building.embark_waiting_passengers(current_tick=5)

        assert len(building.floors[0].waiting_passengers) == 0
        assert len(elevator.onboard_passengers) == 2

    def test_disembark_arrived_passengers(self):
        """
        Given: An elevator with passengers at their destination floor
        When: disembark_arrived_passengers is called
        Then: Passengers leave the elevator and are marked as disembarked
        """
        building = Building(n_floors=5, n_elevators=1)
        elevator = building.elevators[0]

        # Create passenger and manually embark
        passenger = building.floors[0].create_passenger("P__1", building.floors[3], 0)
        building.floors[0].remove_waiting_passenger(passenger)
        passenger.embark(1)
        elevator.onboard_passengers.append(passenger)

        # Move elevator to destination
        elevator.current_floor = building.floors[3]

        # Disembark passengers
        building.disembark_arrived_passengers(current_tick=10)

        assert len(elevator.onboard_passengers) == 0
        assert len(building.floors[3].dismbarked_passengers) == 1

    def test_handle_passengers(self):
        """
        Given: An elevator with passengers arriving and waiting passengers
        When: handle_passengers is called
        Then: Passengers disembark and new passengers embark
        """
        building = Building(n_floors=5, n_elevators=1)
        elevator = building.elevators[0]

        # Add waiting passenger at ground floor
        building.create_and_add_passenger("P__1", "F__0", "F__2", 0)

        # Add onboard passenger arriving at ground floor
        passenger2 = building.floors[2].create_passenger("P__2", building.floors[0], 1)
        building.floors[2].remove_waiting_passenger(passenger2)
        passenger2.embark(2)
        elevator.onboard_passengers.append(passenger2)
        elevator.current_floor = building.floors[0]

        # Handle passengers
        building.handle_passengers(current_tick=5)

        # P__2 should disembark, P__1 should embark
        assert len(building.floors[0].dismbarked_passengers) == 1
        assert len(elevator.onboard_passengers) == 1
        assert elevator.onboard_passengers[0].passenger_id == "P__1"


class TestBuildingIntegration:
    """Integration tests for Building"""

    def test_complete_passenger_journey(self):
        """
        Given: A passenger created and added to a floor
        When: The passenger embarks, elevator moves, and passenger disembarks
        Then: The passenger completes the journey from origin to destination
        """
        building = Building(n_floors=5, n_elevators=1)
        elevator = building.elevators[0]

        # Create passenger
        building.create_and_add_passenger("P__1", "F__0", "F__3", 0)
        assert len(building.floors[0].waiting_passengers) == 1

        # Embark passenger
        building.embark_waiting_passengers(current_tick=1)
        assert len(elevator.onboard_passengers) == 1
        assert len(building.floors[0].waiting_passengers) == 0

        # Move elevator to destination (1 tick to change state + 3 ticks to move)
        elevator.target_floor = building.floors[3]
        for _ in range(4):
            building.tick_movement()

        assert elevator.current_floor == building.floors[3]

        # Disembark passenger
        building.disembark_arrived_passengers(current_tick=5)
        assert len(elevator.onboard_passengers) == 0
        assert len(building.floors[3].dismbarked_passengers) == 1

    def test_multiple_elevators_multiple_passengers(self):
        """
        Given: Multiple elevators and multiple passengers on different floors
        When: embark_waiting_passengers is called
        Then: Passengers are distributed across available elevators
        """
        building = Building(n_floors=10, n_elevators=2, elevator_capacity=5)

        # Create passengers on different floors
        building.create_and_add_passenger("P__1", "F__0", "F__5", 0)
        building.create_and_add_passenger("P__2", "F__0", "F__7", 0)
        building.create_and_add_passenger("P__3", "F__2", "F__8", 0)

        # Embark passengers
        building.embark_waiting_passengers(current_tick=1)

        # Check that at least some passengers embarked
        total_onboard = sum(len(e.onboard_passengers) for e in building.elevators)
        assert total_onboard > 0

    def test_elevator_capacity_limit(self):
        """
        Given: An elevator with limited capacity and many waiting passengers
        When: embark_waiting_passengers is called
        Then: Only capacity-limited number of passengers embark
        """
        building = Building(n_floors=5, n_elevators=1, elevator_capacity=2)
        elevator = building.elevators[0]

        # Create 5 passengers
        for i in range(5):
            building.create_and_add_passenger(f"P__{i}", "F__0", "F__3", 0)

        assert len(building.floors[0].waiting_passengers) == 5

        # Try to embark - should only take 2
        building.embark_waiting_passengers(current_tick=1)

        assert len(elevator.onboard_passengers) == 2
        assert len(building.floors[0].waiting_passengers) == 3



from __future__ import annotations

from typing import Optional, TYPE_CHECKING
from .elevator_logic_base import ElevatorLogic

if TYPE_CHECKING:
    from ..elevator import Elevator
    from ..floor import Floor

class LookLogic(ElevatorLogic):
    def __init__(self, building):
        super().__init__(building)
        # Stores elevator "sweep" direction
        self.elevator_directions = {}

    def __repr__(self) -> str:
        return "LookLogic"

    def assign_elevator_targets(self) -> None:
        # For each elevator, update its target based on LOOK logic
        for elevator in self.building.elevators:
            # If elevator is at target or has no target, find next target
            if elevator.state == 'IDLE' and (elevator.target_floor is None or
                                             elevator.target_floor == elevator.current_floor):
                next_target = self._get_next_target_look(elevator)
                if next_target and next_target != elevator.current_floor:
                    elevator.target_floor = next_target
                elif not next_target and not elevator.onboard_passengers:
                    # No more targets and no passengers, clear target
                    elevator.target_floor = None

            # Even when moving, update target if there's a better stop in the same direction
            elif elevator.state in ['MOVING_UP', 'MOVING_DOWN'] and not elevator.is_at_capacity():
                # Check if there are waiting passengers in the current direction
                next_target = self._get_next_target_look(elevator)
                if next_target and next_target != elevator.current_floor:
                    # Update target if it's in the same direction and closer
                    current_floor_num = int(elevator.current_floor)
                    next_target_num = int(next_target)
                    current_target_num = int(elevator.target_floor) if elevator.target_floor else current_floor_num

                    if elevator.state == 'MOVING_UP':
                        # If new target is between current and old target, or beyond old target in same direction
                        if next_target_num > current_floor_num and next_target_num >= current_target_num:
                            elevator.target_floor = next_target
                    elif elevator.state == 'MOVING_DOWN':
                        # If new target is between current and old target, or beyond old target in same direction
                        if next_target_num < current_floor_num and next_target_num <= current_target_num:
                            elevator.target_floor = next_target

    def _get_next_target_look(self, elevator: Elevator) -> Optional[Floor]:
        current_floor_num = int(elevator.current_floor)

        # Determine direction from elevator state
        if elevator.state == 'MOVING_UP':
            direction = 'UP'
            self.elevator_directions[elevator.elevator_id] = 'UP'
        elif elevator.state == 'MOVING_DOWN':
            direction = 'DOWN'
            self.elevator_directions[elevator.elevator_id] = 'DOWN'
        else:  # IDLE
            # Try to maintain last direction if we have requests in that direction
            last_direction = self.elevator_directions.get(elevator.elevator_id, 'UP')
            direction = last_direction

        # Collect all destinations from onboard passengers
        dest_floors_up = []
        dest_floors_down = []

        for passenger in elevator.onboard_passengers:
            if passenger.destination_floor:
                dest_num = int(passenger.destination_floor)
                if dest_num > current_floor_num:
                    dest_floors_up.append((passenger.destination_floor, dest_num))
                elif dest_num < current_floor_num:
                    dest_floors_down.append((passenger.destination_floor, dest_num))

        # Collect floors with waiting passengers based on their destination direction
        # Crucial for LOOK: only consider passengers whose trip direction we can serve
        if not elevator.is_at_capacity():
            for floor in self.building.floors:
                floor_num = int(floor)
                if floor.waiting_passengers:
                    # Check if any passenger on this floor wants to travel up or down
                    for passenger in floor.waiting_passengers:
                        if passenger.destination_floor:
                            dest_num = int(passenger.destination_floor)
                            # Passenger wants to go UP from their floor
                            if dest_num > floor_num > current_floor_num:
                                dest_floors_up.append((floor, floor_num))
                                break  # Only need to add floor once
                            # Passenger wants to go DOWN from their floor
                            if dest_num < floor_num < current_floor_num:
                                dest_floors_down.append((floor, floor_num))
                                break  # Only need to add floor once

        # LOOK logic: serve all in current direction, then reverse immediately
        result = None

        if direction == 'UP':
            if dest_floors_up:
                # Go to closest floor going up
                dest_floors_up.sort(key=lambda x: x[1])
                result = dest_floors_up[0][0]
            elif dest_floors_down:
                # No more destinations up, reverse immediately
                dest_floors_down.sort(key=lambda x: x[1], reverse=True)
                result = dest_floors_down[0][0]
                # Update direction since we're reversing
                self.elevator_directions[elevator.elevator_id] = 'DOWN'
        else:  # DOWN
            if dest_floors_down:
                # Go to closest floor going down
                dest_floors_down.sort(key=lambda x: x[1], reverse=True)
                result = dest_floors_down[0][0]
            elif dest_floors_up:
                # No more destinations down, reverse immediately
                dest_floors_up.sort(key=lambda x: x[1])
                result = dest_floors_up[0][0]
                # Update direction since we're reversing
                self.elevator_directions[elevator.elevator_id] = 'UP'

        return result



from __future__ import annotations

from typing import Dict, Any, TYPE_CHECKING

from .floor import Floor
from .elevator import Elevator

if TYPE_CHECKING:
    from .passenger import Passenger

class Building:
    def __init__(self, n_floors: int, n_elevators: int, elevator_capacity: int = 10) -> None:
        self.elevator_capacity = elevator_capacity
        self.floors = [Floor(f"F__{i}") for i in range(n_floors)]
        self.elevators = [Elevator(elevator_id=f"E__{i}", building=self, capacity=elevator_capacity) for i in range(n_elevators)]

    def __repr__(self) -> str:
        return f"Building(floors={len(self.floors)} floors, elevators={len(self.elevators)} elevators, elevator_capacity={self.elevator_capacity})"

    def to_dict(self) -> Dict[str, Any]:
        return {
            'type': 'Building',
            'elevator_capacity': self.elevator_capacity,
            'floors': {f"F__{i}": floor.to_dict() for i, floor in enumerate(self.floors)},
            'elevators': {f"E__{i}": elevator.to_dict() for i, elevator in enumerate(self.elevators)},
        }

    def tick_movement(self) -> None:
        for elevator in self.elevators:
            elevator.tick_movement()

    def handle_passengers(self, current_tick: int) -> None:
        self.disembark_arrived_passengers(current_tick)
        self.embark_waiting_passengers(current_tick)

    def embark_waiting_passengers(self, current_tick: int) -> None:
        for elevator in self.elevators:
            elevator.embark_waiting_passengers(current_tick)

    def disembark_arrived_passengers(self, current_tick: int) -> None:
        for elevator in self.elevators:
            elevator.disembark_arrived_passengers(current_tick)

    def add_waiting_passenger(self, passenger: Passenger) -> None:
        passenger.origin_floor.add_waiting_passenger(passenger)

    def create_and_add_passenger(self, passenger_id: str, origin_floor_id: str, destination_floor_id: str, tick_appeared: int) -> bool:
        try:
            origin_floor_num = int(origin_floor_id.split('__')[1])
            destination_floor_num = int(destination_floor_id.split('__')[1])

            if origin_floor_num < 0 or origin_floor_num >= len(self.floors):
                print(f"Warning: Origin floor {origin_floor_id} does not exist.")
                return False

            if destination_floor_num < 0 or destination_floor_num >= len(self.floors):
                print(f"Warning: Destination floor {destination_floor_id} does not exist.")
                return False

            origin_floor = self.floors[origin_floor_num]
            destination_floor = self.floors[destination_floor_num]

            origin_floor.create_passenger(passenger_id, destination_floor, tick_appeared)
            return True

        except (ValueError, IndexError) as e:
            print(f"Warning: Could not create passenger {passenger_id}: {e}")
            return False



from setuptools import setup, find_packages
import os
import re

# Read version from __init__.py
def get_version():
    init_path = os.path.join(os.path.dirname(__file__), "elevator_traffic_analysis", "__init__.py")
    with open(init_path, "r", encoding="utf-8") as f:
        content = f.read()
        match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]", content, re.MULTILINE)
        if match:
            return match.group(1)
        raise RuntimeError("Unable to find version string.")

try:
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
except FileNotFoundError:
    long_description = "A Python package for elevator traffic analysis and simulation"

try:
    with open("requirements.txt", "r", encoding="utf-8") as fh:
        requirements = [line.strip() for line in fh if line.strip() and not line.startswith("#")]
except FileNotFoundError:
    requirements = []

setup(
    name="elevator-traffic-analysis",
    version=get_version(),
    author="tkosman",
    author_email="tymoteusz4development@gmail.com",
    description="A Python package for elevator traffic analysis and simulation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tkosman/elevator-traffic-analysis",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.8",
    install_requires=requirements,
    extras_require={
        "dev": ["pytest", "tox"],
    },
    entry_points={
        "console_scripts": [
            "elevator-traffic-analysis=elevator_traffic_analysis.__main__:main",
        ],
    },
    include_package_data=True,
    package_data={
        "elevator_traffic_analysis": [],
    },
    data_files=[
        ("scenarios", ["scenarios/*.yaml"]),
    ],
)


from elevator_traffic_analysis.floor import Floor
from elevator_traffic_analysis.passenger import Passenger


def test_floor_creation():
    """
    Given: A floor ID is provided
    When: A Floor object is created
    Then: The floor has correct properties and empty passenger lists
    """
    floor = Floor(floor_id="floor__1")
    assert floor.floor_id == "floor__1"
    assert len(floor.waiting_passengers) == 0
    assert len(floor.dismbarked_passengers) == 0
    assert int(floor) == 1


def test_passenger_creation():
    """
    Given: Origin and destination floors exist
    When: A passenger is created via floor's create_passenger method
    Then: The passenger is initialized correctly and added to waiting passengers
    """
    origin_floor = Floor(floor_id="floor__0")
    destination_floor = Floor(floor_id="floor__5")

    passenger = origin_floor.create_passenger(
        passenger_id="passenger_1",
        destination_floor=destination_floor,
        tick_appeared=10
    )

    assert passenger.passenger_id == "passenger_1"
    assert passenger.origin_floor == origin_floor
    assert passenger.destination_floor == destination_floor
    assert passenger.tick_appeared == 10
    assert passenger.state == "WAITING"
    assert len(origin_floor.waiting_passengers) == 1


def test_passenger_state_transitions():
    """
    Given: A passenger in WAITING state
    When: The passenger embarks and then disembarks
    Then: The passenger state transitions correctly through all stages
    """
    origin_floor = Floor(floor_id="floor__0")
    destination_floor = Floor(floor_id="floor__5")

    passenger = Passenger(
        passenger_id="passenger_1",
        origin_floor=origin_floor,
        destination_floor=destination_floor,
        tick_appeared=10
    )

    assert passenger.state == "WAITING"

    passenger.embark(tick=15)
    assert passenger.state == "EMBARKED"
    assert passenger.tick_embarked == 15

    passenger.disembark(tick=25)
    assert passenger.state == "DISEMBARKED"
    assert passenger.tick_disembarked == 25


def test_passenger_stats():
    """
    Given: A passenger that has completed their journey
    When: get_stats is called
    Then: Wait, ride, and serve time statistics are correctly calculated
    """
    origin_floor = Floor(floor_id="floor__0")
    destination_floor = Floor(floor_id="floor__5")

    passenger = Passenger(
        passenger_id="passenger_1",
        origin_floor=origin_floor,
        destination_floor=destination_floor,
        tick_appeared=10
    )

    # Stats should be None before disembarking
    assert passenger.get_stats() is None

    passenger.embark(tick=15)
    assert passenger.get_stats() is None

    passenger.disembark(tick=25)
    stats = passenger.get_stats()

    assert stats is not None
    assert stats["tick_appeared"] == 10
    assert stats["tick_embarked"] == 15
    assert stats["tick_disembarked"] == 25
    assert stats["n_wait_ticks"] == 5  # 15 - 10
    assert stats["n_ride_ticks"] == 10  # 25 - 15
    assert stats["n_serve_ticks"] == 15  # 25 - 10



import pytest
from elevator_traffic_analysis.building import Building
from elevator_traffic_analysis.elevator_algorithms.fcfs import FirstComeFirstServedLogic
from elevator_traffic_analysis.elevator_algorithms.sstf import ShortestSeekTimeFirstLogic
from elevator_traffic_analysis.elevator_algorithms.scan import ScanLogic
from elevator_traffic_analysis.elevator_algorithms.look import LookLogic


class TestFCFSLogic:
    """Tests for First Come First Served (FCFS) algorithm"""

    def test_fcfs_creation(self):
        """
        Given: A building with floors and elevators
        When: FCFS logic is created
        Then: The logic is properly initialized with building reference
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = FirstComeFirstServedLogic(building)
        assert logic.building is building
        assert repr(logic) == "FirstComeFirstServedLogic"

    def test_fcfs_no_passengers_no_target(self):
        """
        Given: A building with no passengers
        When: FCFS assigns elevator targets
        Then: No target is set for the elevator
        """
        building = Building(n_floors=5, n_elevators=1)
        logic = FirstComeFirstServedLogic(building)
        elevator = building.elevators[0]

        logic.assign_elevator_targets()

        assert elevator.target_floor is None

    def test_fcfs_onboard_passenger_priority(self):
        """
        Given: An elevator with an onboard passenger
        When: FCFS assigns elevator targets
        Then: The onboard passenger's destination is prioritized
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = FirstComeFirstServedLogic(building)
        elevator = building.elevators[0]

        # Add onboard passenger going to floor 5
        passenger = building.floors[0].create_passenger("P__1", building.floors[5], 0)
        building.floors[0].remove_waiting_passenger(passenger)
        passenger.embark(1)
        elevator.onboard_passengers.append(passenger)

        logic.assign_elevator_targets()

        assert elevator.target_floor == building.floors[5]

    def test_fcfs_picks_earliest_waiting_passenger(self):
        """
        Given: Multiple waiting passengers with different appearance times
        When: FCFS assigns elevator targets
        Then: The earliest waiting passenger's floor is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = FirstComeFirstServedLogic(building)
        elevator = building.elevators[0]

        # Add passengers at different times
        building.create_and_add_passenger("P__1", "F__2", "F__7", tick_appeared=10)
        building.create_and_add_passenger("P__2", "F__3", "F__8", tick_appeared=5)
        building.create_and_add_passenger("P__3", "F__4", "F__9", tick_appeared=15)

        logic.assign_elevator_targets()

        # Should pick floor 3 (P__2 appeared at tick 5, earliest)
        assert elevator.target_floor == building.floors[3]

    def test_fcfs_first_embarked_passenger_priority(self):
        """
        Given: Multiple onboard passengers who embarked at different times
        When: FCFS assigns elevator targets
        Then: The first embarked passenger's destination is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = FirstComeFirstServedLogic(building)
        elevator = building.elevators[0]

        # Add multiple onboard passengers
        p1 = building.floors[0].create_passenger("P__1", building.floors[7], 0)
        p2 = building.floors[0].create_passenger("P__2", building.floors[5], 0)
        p3 = building.floors[0].create_passenger("P__3", building.floors[9], 0)

        building.floors[0].remove_waiting_passenger(p1)
        building.floors[0].remove_waiting_passenger(p2)
        building.floors[0].remove_waiting_passenger(p3)

        p1.embark(10)
        p2.embark(5)
        p3.embark(15)

        elevator.onboard_passengers.extend([p1, p2, p3])

        logic.assign_elevator_targets()

        # Should pick p2's destination (embarked at tick 5)
        assert elevator.target_floor == building.floors[5]

    def test_fcfs_ignores_at_capacity(self):
        """
        Given: An elevator at capacity with waiting passengers
        When: FCFS assigns elevator targets
        Then: Only onboard passenger destinations are considered
        """
        building = Building(n_floors=10, n_elevators=1, elevator_capacity=2)
        logic = FirstComeFirstServedLogic(building)
        elevator = building.elevators[0]

        # Fill elevator to capacity
        p1 = building.floors[0].create_passenger("P__1", building.floors[5], 0)
        p2 = building.floors[0].create_passenger("P__2", building.floors[6], 0)
        building.floors[0].remove_waiting_passenger(p1)
        building.floors[0].remove_waiting_passenger(p2)
        p1.embark(1)
        p2.embark(2)
        elevator.onboard_passengers.extend([p1, p2])

        # Add waiting passenger
        building.create_and_add_passenger("P__3", "F__3", "F__8", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should target onboard passenger destination, not waiting passenger
        assert elevator.target_floor in [building.floors[5], building.floors[6]]

    def test_fcfs_clears_target_when_no_passengers(self):
        """
        Given: An elevator at target floor with no passengers
        When: FCFS assigns elevator targets
        Then: The target is cleared
        """
        building = Building(n_floors=5, n_elevators=1)
        logic = FirstComeFirstServedLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[3]
        elevator.target_floor = building.floors[3]

        logic.assign_elevator_targets()

        assert elevator.target_floor is None

    def test_fcfs_doesnt_update_while_moving(self):
        """
        Given: An elevator moving towards a target with new passengers appearing
        When: FCFS assigns elevator targets
        Then: The target remains unchanged
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = FirstComeFirstServedLogic(building)
        elevator = building.elevators[0]

        # Set target and start moving
        elevator.target_floor = building.floors[5]
        elevator.move_up()

        # Add new passenger
        building.create_and_add_passenger("P__1", "F__2", "F__7", tick_appeared=0)

        logic.assign_elevator_targets()

        # Target should remain floor 5
        assert elevator.target_floor == building.floors[5]


class TestSSTFLogic:
    """Tests for Shortest Seek Time First (SSTF) algorithm"""

    def test_sstf_creation(self):
        """
        Given: A building with floors and elevators
        When: SSTF logic is created
        Then: The logic is properly initialized with building reference
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ShortestSeekTimeFirstLogic(building)
        assert logic.building is building
        assert repr(logic) == "ShortestSeekTimeFirstLogic"

    def test_sstf_picks_closest_destination(self):
        """
        Given: Multiple passengers at different distances from elevator
        When: SSTF assigns elevator targets
        Then: The closest passenger's floor is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ShortestSeekTimeFirstLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]

        # Add passengers at different distances
        building.create_and_add_passenger("P__1", "F__2", "F__8", tick_appeared=0)  # Distance 3
        building.create_and_add_passenger("P__2", "F__7", "F__9", tick_appeared=0)  # Distance 2
        building.create_and_add_passenger("P__3", "F__9", "F__1", tick_appeared=0)  # Distance 4

        logic.assign_elevator_targets()

        # Should pick floor 7 (closest at distance 2)
        assert elevator.target_floor == building.floors[7]

    def test_sstf_onboard_passenger_closest(self):
        """
        Given: Multiple onboard passengers with different destination distances
        When: SSTF assigns elevator targets
        Then: The closest destination is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ShortestSeekTimeFirstLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]

        # Add onboard passengers
        p1 = building.floors[0].create_passenger("P__1", building.floors[9], 0)
        p2 = building.floors[0].create_passenger("P__2", building.floors[6], 0)
        p3 = building.floors[0].create_passenger("P__3", building.floors[3], 0)

        building.floors[0].remove_waiting_passenger(p1)
        building.floors[0].remove_waiting_passenger(p2)
        building.floors[0].remove_waiting_passenger(p3)

        p1.embark(1)
        p2.embark(1)
        p3.embark(1)
        elevator.onboard_passengers.extend([p1, p2, p3])

        logic.assign_elevator_targets()

        # Should pick floor 6 (distance 1)
        assert elevator.target_floor == building.floors[6]

    def test_sstf_can_switch_targets_while_moving(self):
        """
        Given: An elevator moving to a target when a closer passenger appears
        When: SSTF assigns elevator targets
        Then: The target switches to the closer floor
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ShortestSeekTimeFirstLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[0]
        elevator.target_floor = building.floors[8]
        elevator.move_up()

        # Add closer waiting passenger
        building.create_and_add_passenger("P__1", "F__3", "F__7", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should switch to floor 3 (closer than floor 8)
        assert elevator.target_floor == building.floors[3]

    def test_sstf_excludes_current_floor(self):
        """
        Given: Passengers at current floor and other floors
        When: SSTF assigns elevator targets
        Then: Current floor is excluded and next closest floor is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ShortestSeekTimeFirstLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]

        # Add passengers: one at current floor, one elsewhere
        building.create_and_add_passenger("P__1", "F__5", "F__8", tick_appeared=0)
        building.create_and_add_passenger("P__2", "F__7", "F__9", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should target floor 7 (closer than 8 and not current floor)
        assert elevator.target_floor == building.floors[7]

    def test_sstf_no_passengers_no_target(self):
        """
        Given: A building with no passengers
        When: SSTF assigns elevator targets
        Then: No target is set for the elevator
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ShortestSeekTimeFirstLogic(building)
        elevator = building.elevators[0]

        logic.assign_elevator_targets()

        assert elevator.target_floor is None

    def test_sstf_respects_capacity(self):
        """
        Given: An elevator at capacity with closer waiting passengers
        When: SSTF assigns elevator targets
        Then: Only onboard passenger destinations are considered
        """
        building = Building(n_floors=10, n_elevators=1, elevator_capacity=1)
        logic = ShortestSeekTimeFirstLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]

        # Fill elevator
        p1 = building.floors[0].create_passenger("P__1", building.floors[9], 0)
        building.floors[0].remove_waiting_passenger(p1)
        p1.embark(1)
        elevator.onboard_passengers.append(p1)

        # Add closer waiting passenger
        building.create_and_add_passenger("P__2", "F__6", "F__8", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should target floor 9 (onboard passenger), not floor 6
        assert elevator.target_floor == building.floors[9]


class TestScanLogic:
    """Tests for SCAN (Elevator) algorithm"""

    def test_scan_creation(self):
        """
        Given: A building with floors and elevators
        When: SCAN logic is created
        Then: The logic is properly initialized with building reference
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        assert logic.building is building
        assert repr(logic) == "ScanLogic"

    def test_scan_continues_direction_up(self):
        """
        Given: An elevator moving upward with an upward request
        When: SCAN assigns elevator targets
        Then: The elevator continues to target upward floors
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[3]

        # Add passenger going up
        p1 = building.floors[0].create_passenger("P__1", building.floors[7], 0)
        building.floors[0].remove_waiting_passenger(p1)
        p1.embark(1)
        elevator.onboard_passengers.append(p1)

        # Set elevator moving up
        elevator.target_floor = building.floors[7]
        elevator.move_up()
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        logic.assign_elevator_targets()

        # Should continue to floor 7 or beyond
        assert int(elevator.target_floor) >= 7

    def test_scan_continues_direction_down(self):
        """
        Given: An elevator moving downward with a downward request
        When: SCAN assigns elevator targets
        Then: The elevator continues to target downward floors
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[7]

        # Add passenger going down
        p1 = building.floors[8].create_passenger("P__1", building.floors[2], 0)
        building.floors[8].remove_waiting_passenger(p1)
        p1.embark(1)
        elevator.onboard_passengers.append(p1)

        # Set elevator moving down
        elevator.target_floor = building.floors[2]
        elevator.move_down()
        logic.elevator_directions[elevator.elevator_id] = 'DOWN'

        logic.assign_elevator_targets()

        # Should continue to floor 2 or below
        assert int(elevator.target_floor) <= 2

    def test_scan_goes_to_boundary_when_no_requests(self):
        """
        Given: An elevator with no requests in current direction
        When: SCAN assigns elevator targets
        Then: The elevator targets the boundary floor
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        logic.assign_elevator_targets()

        # Should go to top floor (9)
        assert elevator.target_floor == building.floors[9]

    def test_scan_reverses_at_top(self):
        """
        Given: An elevator at top boundary with downward requests
        When: SCAN assigns elevator targets
        Then: The elevator reverses direction and targets downward
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[9]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add passenger going down
        building.create_and_add_passenger("P__1", "F__5", "F__2", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should reverse and target floor 5
        assert elevator.target_floor == building.floors[5]

    def test_scan_reverses_at_bottom(self):
        """
        Given: An elevator at bottom boundary with upward requests
        When: SCAN assigns elevator targets
        Then: The elevator reverses direction and targets upward
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[0]
        logic.elevator_directions[elevator.elevator_id] = 'DOWN'

        # Add passenger going up
        building.create_and_add_passenger("P__1", "F__5", "F__8", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should reverse and target floor 5
        assert elevator.target_floor == building.floors[5]

    def test_scan_picks_furthest_in_direction(self):
        """
        Given: Multiple passengers in elevator's current direction
        When: SCAN assigns elevator targets
        Then: The furthest floor in that direction is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[2]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add multiple passengers going up
        building.create_and_add_passenger("P__1", "F__4", "F__6", tick_appeared=0)
        building.create_and_add_passenger("P__2", "F__6", "F__8", tick_appeared=0)
        building.create_and_add_passenger("P__3", "F__3", "F__5", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should pick furthest (floor 6)
        assert int(elevator.target_floor) >= 6

    def test_scan_ignores_opposite_direction_passengers(self):
        """
        Given: Passengers in both directions with elevator moving up
        When: SCAN assigns elevator targets
        Then: Only upward passengers are considered
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add passengers: one going up from above, one going down from below
        building.create_and_add_passenger("P__1", "F__2", "F__0", tick_appeared=0)  # Below, going down
        building.create_and_add_passenger("P__2", "F__7", "F__9", tick_appeared=0)  # Above, going up

        logic.assign_elevator_targets()

        # Should pick floor 7 (in current direction), not floor 2
        assert int(elevator.target_floor) >= 7

    def test_scan_default_direction_up(self):
        """
        Given: An elevator without a set direction and passengers in both directions
        When: SCAN assigns elevator targets
        Then: The elevator defaults to moving upward
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = ScanLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]

        # Add passengers both above and below
        building.create_and_add_passenger("P__1", "F__2", "F__0", tick_appeared=0)
        building.create_and_add_passenger("P__2", "F__7", "F__9", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should go up by default
        assert int(elevator.target_floor) >= 5


class TestLookLogic:
    """Tests for LOOK (C-LOOK variant) algorithm"""

    def test_look_creation(self):
        """
        Given: A building with floors and elevators
        When: LOOK logic is created
        Then: The logic is properly initialized with building reference
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        assert logic.building is building
        assert repr(logic) == "LookLogic"

    def test_look_continues_direction_up(self):
        """
        Given: An elevator moving upward with an upward request
        When: LOOK assigns elevator targets
        Then: The elevator continues to target upward floors
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[3]

        # Add passenger going up
        p1 = building.floors[0].create_passenger("P__1", building.floors[7], 0)
        building.floors[0].remove_waiting_passenger(p1)
        p1.embark(1)
        elevator.onboard_passengers.append(p1)

        # Set elevator moving up
        elevator.target_floor = building.floors[7]
        elevator.move_up()
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        logic.assign_elevator_targets()

        # Should continue to floor 7
        assert elevator.target_floor == building.floors[7]

    def test_look_reverses_immediately_when_no_requests(self):
        """
        Given: An elevator with no requests in current direction but requests behind
        When: LOOK assigns elevator targets
        Then: The elevator immediately reverses direction
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add passenger below current floor going down
        building.create_and_add_passenger("P__1", "F__2", "F__0", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should reverse immediately to floor 2
        assert elevator.target_floor == building.floors[2]

    def test_look_picks_closest_in_direction(self):
        """
        Given: Multiple passengers in elevator's current direction
        When: LOOK assigns elevator targets
        Then: The closest floor in that direction is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[2]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add multiple passengers going up
        building.create_and_add_passenger("P__1", "F__4", "F__6", tick_appeared=0)
        building.create_and_add_passenger("P__2", "F__7", "F__9", tick_appeared=0)
        building.create_and_add_passenger("P__3", "F__3", "F__5", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should pick closest (floor 3)
        assert elevator.target_floor == building.floors[3]

    def test_look_doesnt_go_to_boundary(self):
        """
        Given: An elevator with requests only in opposite direction
        When: LOOK assigns elevator targets
        Then: The elevator reverses without reaching boundary
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add passenger below going down
        building.create_and_add_passenger("P__1", "F__3", "F__1", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should reverse to floor 3, not go to top boundary first
        assert elevator.target_floor == building.floors[3]
        assert int(elevator.target_floor) < 9

    def test_look_serves_onboard_first(self):
        """
        Given: Multiple onboard passengers in current direction
        When: LOOK assigns elevator targets
        Then: The closest onboard destination is targeted
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[2]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add onboard passengers
        p1 = building.floors[0].create_passenger("P__1", building.floors[5], 0)
        p2 = building.floors[0].create_passenger("P__2", building.floors[7], 0)
        building.floors[0].remove_waiting_passenger(p1)
        building.floors[0].remove_waiting_passenger(p2)
        p1.embark(1)
        p2.embark(1)
        elevator.onboard_passengers.extend([p1, p2])

        logic.assign_elevator_targets()

        # Should target floor 5 (closest onboard destination)
        assert elevator.target_floor == building.floors[5]

    def test_look_reverses_direction_tracking(self):
        """
        Given: An elevator that needs to reverse direction
        When: LOOK assigns elevator targets
        Then: The direction tracking is updated accordingly
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[7]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add passenger below going down
        building.create_and_add_passenger("P__1", "F__3", "F__1", tick_appeared=0)

        logic.assign_elevator_targets()

        # Direction should be updated to DOWN
        assert logic.elevator_directions[elevator.elevator_id] == 'DOWN'

    def test_look_ignores_opposite_direction_passengers(self):
        """
        Given: Passengers in both directions with elevator moving up
        When: LOOK assigns elevator targets
        Then: Only upward passengers are considered
        """
        building = Building(n_floors=10, n_elevators=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[5]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Add passengers: one above going up, one below going down
        building.create_and_add_passenger("P__1", "F__2", "F__0", tick_appeared=0)  # Below, going down
        building.create_and_add_passenger("P__2", "F__7", "F__9", tick_appeared=0)  # Above, going up

        logic.assign_elevator_targets()

        # Should pick floor 7 (in current direction)
        assert elevator.target_floor == building.floors[7]

    def test_look_respects_capacity(self):
        """
        Given: An elevator at capacity with closer waiting passengers
        When: LOOK assigns elevator targets
        Then: Only onboard passenger destinations are considered
        """
        building = Building(n_floors=10, n_elevators=1, elevator_capacity=1)
        logic = LookLogic(building)
        elevator = building.elevators[0]
        elevator.current_floor = building.floors[3]
        logic.elevator_directions[elevator.elevator_id] = 'UP'

        # Fill elevator
        p1 = building.floors[0].create_passenger("P__1", building.floors[8], 0)
        building.floors[0].remove_waiting_passenger(p1)
        p1.embark(1)
        elevator.onboard_passengers.append(p1)

        # Add waiting passenger closer
        building.create_and_add_passenger("P__2", "F__5", "F__7", tick_appeared=0)

        logic.assign_elevator_targets()

        # Should target floor 8 (onboard), not floor 5 (waiting)
        assert elevator.target_floor == building.floors[8]


class TestAlgorithmComparison:
    """Integration tests comparing algorithm behaviors"""

    def test_all_algorithms_handle_simple_case(self):
        """
        Given: All algorithms with a single passenger
        When: Each algorithm assigns elevator targets
        Then: All algorithms set a target floor
        """
        algorithms_with_buildings = [
            (FirstComeFirstServedLogic, Building(n_floors=10, n_elevators=1)),
            (ShortestSeekTimeFirstLogic, Building(n_floors=10, n_elevators=1)),
            (ScanLogic, Building(n_floors=10, n_elevators=1)),
            (LookLogic, Building(n_floors=10, n_elevators=1))
        ]

        for AlgoClass, building in algorithms_with_buildings:
            # Add passenger at floor 2 (not ground floor where elevator starts)
            building.create_and_add_passenger("P__1", "F__2", "F__5", tick_appeared=0)
            logic = AlgoClass(building)
            elevator = building.elevators[0]
            elevator.target_floor = None

            logic.assign_elevator_targets()

            # All should set a target (either floor 2 to pick up or go somewhere)
            assert elevator.target_floor is not None

    def test_fcfs_and_sstf_clear_target_when_done(self):
        """
        Given: FCFS and SSTF elevators at target with no passengers
        When: Each algorithm assigns elevator targets
        Then: Both algorithms clear the target
        """
        # Only test FCFS and SSTF as SCAN/LOOK may set targets to boundaries
        algorithms_with_buildings = [
            (FirstComeFirstServedLogic, Building(n_floors=5, n_elevators=1)),
            (ShortestSeekTimeFirstLogic, Building(n_floors=5, n_elevators=1))
        ]

        for AlgoClass, building in algorithms_with_buildings:
            logic = AlgoClass(building)
            elevator = building.elevators[0]
            # Set elevator at target floor with no passengers
            elevator.current_floor = building.floors[3]
            elevator.target_floor = building.floors[3]

            logic.assign_elevator_targets()

            # FCFS and SSTF should clear target when no passengers and at target
            assert elevator.target_floor is None

    def test_fcfs_vs_sstf_different_choices(self):
        """
        Given: Same scenario with close/late and far/early passengers
        When: FCFS and SSTF assign elevator targets
        Then: FCFS picks earliest, SSTF picks closest
        """
        building_fcfs = Building(n_floors=10, n_elevators=1)
        building_sstf = Building(n_floors=10, n_elevators=1)

        # Setup: elevator at floor 5, passengers at different distances and times
        building_fcfs.elevators[0].current_floor = building_fcfs.floors[5]
        building_sstf.elevators[0].current_floor = building_sstf.floors[5]

        # Close passenger (distance 1) but appeared later (tick 10)
        building_fcfs.create_and_add_passenger("P__1", "F__6", "F__8", tick_appeared=10)
        building_sstf.create_and_add_passenger("P__1", "F__6", "F__8", tick_appeared=10)

        # Far passenger (distance 5) but appeared earlier (tick 5)
        building_fcfs.create_and_add_passenger("P__2", "F__0", "F__3", tick_appeared=5)
        building_sstf.create_and_add_passenger("P__2", "F__0", "F__3", tick_appeared=5)

        fcfs_logic = FirstComeFirstServedLogic(building_fcfs)
        sstf_logic = ShortestSeekTimeFirstLogic(building_sstf)

        fcfs_logic.assign_elevator_targets()
        sstf_logic.assign_elevator_targets()

        # FCFS should pick floor 0 (earlier time)
        assert building_fcfs.elevators[0].target_floor == building_fcfs.floors[0]

        # SSTF should pick floor 6 (closer distance)
        assert building_sstf.elevators[0].target_floor == building_sstf.floors[6]

    def test_scan_vs_look_boundary_behavior(self):
        """
        Given: Same scenario with no requests ahead but requests behind
        When: SCAN and LOOK assign elevator targets
        Then: SCAN goes to boundary, LOOK reverses immediately
        """
        building_scan = Building(n_floors=10, n_elevators=1)
        building_look = Building(n_floors=10, n_elevators=1)

        # Setup: elevator at floor 5, moving up, no passengers above
        building_scan.elevators[0].current_floor = building_scan.floors[5]
        building_look.elevators[0].current_floor = building_look.floors[5]

        scan_logic = ScanLogic(building_scan)
        look_logic = LookLogic(building_look)

        scan_logic.elevator_directions[building_scan.elevators[0].elevator_id] = 'UP'
        look_logic.elevator_directions[building_look.elevators[0].elevator_id] = 'UP'

        # Add passenger below going down
        building_scan.create_and_add_passenger("P__1", "F__2", "F__0", tick_appeared=0)
        building_look.create_and_add_passenger("P__1", "F__2", "F__0", tick_appeared=0)

        scan_logic.assign_elevator_targets()
        look_logic.assign_elevator_targets()

        # SCAN should go to top boundary (floor 9)
        assert building_scan.elevators[0].target_floor == building_scan.floors[9]

        # LOOK should reverse immediately to floor 2
        assert building_look.elevators[0].target_floor == building_look.floors[2]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional

from transitions import Machine

if TYPE_CHECKING:
    from .floor import Floor

class Passenger:
    states = ["WAITING", "EMBARKED", "DISEMBARKED"]

    def __init__(self, passenger_id: str, origin_floor: Floor, destination_floor: Floor, tick_appeared: int) -> None:
        self.passenger_id: str = passenger_id
        self.origin_floor: Floor = origin_floor
        self.destination_floor: Floor = destination_floor

        self.machine = Machine(model=self, states=Passenger.states, initial="WAITING")
        self.machine.add_transition("_embark", "WAITING", "EMBARKED")
        self.machine.add_transition("_disembark", "EMBARKED", "DISEMBARKED")

        self.tick_appeared: int = tick_appeared
        self.tick_embarked: Optional[int] = None
        self.tick_disembarked: Optional[int] = None

    def embark(self, tick: int) -> None:
        self._embark()
        self.tick_embarked = tick

    def disembark(self, tick: int) -> None:
        self._disembark()
        self.tick_disembarked = tick

    def __repr__(self) -> str:
        return f"Passenger(id={self.passenger_id}, origin={self.origin_floor.floor_id}, destination={self.destination_floor.floor_id}, state={self.state})"

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation for pretty printing"""
        return {
            'type': 'Passenger',
            'id': self.passenger_id,
            'origin_floor_id': self.origin_floor.floor_id,
            'destination_floor_id': self.destination_floor.floor_id,
            'state': self.state,
            'tick_appeared': self.tick_appeared,
            'tick_embarked': self.tick_embarked,
            'tick_disembarked': self.tick_disembarked,
            'stats': self.get_stats()
        }


    def get_stats(self) -> Optional[Dict[str, int]]:
        if self.state != "DISEMBARKED":
            return None
        return {
            "tick_appeared": self.tick_appeared,
            "tick_embarked": self.tick_embarked,
            "tick_disembarked": self.tick_disembarked,
            "n_serve_ticks": self.tick_disembarked - self.tick_appeared,
            "n_ride_ticks": self.tick_disembarked - self.tick_embarked,
            "n_wait_ticks": self.tick_embarked - self.tick_appeared
        }



from __future__ import annotations

from functools import total_ordering
from typing import List, Dict, Any

from .passenger import Passenger

@total_ordering
class Floor:
    def __init__(self, floor_id: str) -> None:
        self.floor_id: str = floor_id
        self.waiting_passengers: List[Passenger] = []
        self.dismbarked_passengers: List[Passenger] = []

    def __lt__(self, other: Floor) -> bool:
        return int(self) < int(other)

    def __eq__(self, other: Floor) -> bool:
        return int(self) == int(other)

    def __int__(self) -> int:
        return int(self.floor_id.split('__')[1])

    def __repr__(self) -> str:
        return f"Floor(id={self.floor_id}, waiting_passengers={len(self.waiting_passengers)})"

    def to_dict(self) -> Dict[str, Any]:
        return {
            'type': 'Floor',
            'id': self.floor_id,
            'waiting_passengers_count': len(self.waiting_passengers),
            'waiting_passengers': [p.to_dict() for p in self.waiting_passengers],
            'dismbarked_passengers_count': len(self.dismbarked_passengers),
            'dismbarked_passengers': [p.to_dict() for p in self.dismbarked_passengers]
        }

    def add_waiting_passenger(self, passenger: Passenger) -> None:
        self.waiting_passengers.append(passenger)

    def remove_waiting_passenger(self, passenger: Passenger) -> None:
        self.waiting_passengers.remove(passenger)

    def add_dismbarked_passenger(self, passenger: Passenger) -> None:
        self.dismbarked_passengers.append(passenger)

    def create_passenger(self, passenger_id: str, destination_floor: Floor, tick_appeared: int) -> Passenger:
        passenger = Passenger(
            passenger_id=passenger_id,
            origin_floor=self,
            destination_floor=destination_floor,
            tick_appeared=tick_appeared
        )
        self.add_waiting_passenger(passenger)
        return passenger



from __future__ import annotations

from typing import Optional, TYPE_CHECKING
from .elevator_logic_base import ElevatorLogic

if TYPE_CHECKING:
    from ..elevator import Elevator
    from ..floor import Floor

class ShortestSeekTimeFirstLogic(ElevatorLogic):

    def __repr__(self) -> str:
        return "ShortestSeekTimeFirstLogic"

    def assign_elevator_targets(self) -> None:
        # For each elevator, update its target based on SSTF logic
        for elevator in self.building.elevators:
            # If elevator is at target or has no target, find next target
            if elevator.state == 'IDLE' and (elevator.target_floor is None or
                                             elevator.target_floor == elevator.current_floor):
                next_target = self._get_next_target_sstf(elevator)
                if next_target and next_target != elevator.current_floor:
                    elevator.target_floor = next_target
                elif not next_target and not elevator.onboard_passengers:
                    # No more targets and no passengers, clear target
                    elevator.target_floor = None

            # Even when moving, update target if a closer floor appears
            elif elevator.state in ['MOVING_UP', 'MOVING_DOWN']:
                next_target = self._get_next_target_sstf(elevator)
                if next_target and next_target != elevator.current_floor:
                    # Update to closer target (SSTF always picks shortest distance)
                    current_target_num = int(elevator.target_floor) if elevator.target_floor else float('inf')
                    next_target_num = int(next_target)
                    current_floor_num = int(elevator.current_floor)

                    # Calculate distances
                    current_target_distance = abs(current_floor_num - current_target_num)
                    next_target_distance = abs(current_floor_num - next_target_num)

                    # If needed, switch to closer target
                    if next_target_distance < current_target_distance:
                        elevator.target_floor = next_target

    def _get_next_target_sstf(self, elevator: Elevator) -> Optional[Floor]:
        current_floor_num = int(elevator.current_floor)

        closest_floor = None
        min_distance = float('inf')

        # Check onboard passenger destinations
        for passenger in elevator.onboard_passengers:
            if passenger.destination_floor:
                dest_floor_num = int(passenger.destination_floor)
                distance = abs(current_floor_num - dest_floor_num)
                if 0 < distance < min_distance:  # Exclude current floor
                    min_distance = distance
                    closest_floor = passenger.destination_floor

        # Check floors with waiting passengers (if elevator has capacity)
        if not elevator.is_at_capacity():
            for floor in self.building.floors:
                if floor.waiting_passengers:
                    floor_num = int(floor)
                    distance = abs(current_floor_num - floor_num)
                    if 0 < distance < min_distance:  # Exclude current floor
                        min_distance = distance
                        closest_floor = floor

        return closest_floor
